# no-op

from .route import route
from .http_methods import *
from .route_holder import RouteHolder
from .request_handler import RequestHandler
from .response_handler import ResponseHandler
from .validations import *
from .jsonapi import jsonapi


import json
import re
from functools import partial
from .singleton import Singleton
from .route_parser import parse_route_for_path
from .enhanced_request import EnhancedRequest

class RouteHolder(metaclass=Singleton):
  def __init__(self):
    self._routes = {}

  def _hoist_request(self, route, event, *, respondent):
    enhanced_request = EnhancedRequest(event, route)
    for before_action in route['before_actions']:
      before_action(enhanced_request, respondent=respondent)
      if respondent.return_immediately:
        break

    if respondent.return_immediately or respondent.has_errors:
      return

    return route['function'](enhanced_request, respondent=respondent)

  def _evaluate_route(self, path, http_method, route):
    if route['http_method'] != 'ANY' and route['http_method'] != http_method:
      return False
    
    if parse_route_for_path(route['route'], path) is None:
      return False
    return True

  def find_best_route(self, path, http_method):
    path_evaluator = partial(self._evaluate_route, path, http_method)
    valid_routes = sorted(filter(path_evaluator, self.routes), key=lambda x: -x['route'].count('/'))
    if valid_routes:
      return partial(self._hoist_request, valid_routes[0])
    return None

  def add_route_partial(self, func, route=None, http_method=None, jsonapi=None, append_before_action=None):
    key = '{}-{}'.format(func.__module__, func.__name__)
    if not key in self._routes:
      self._routes[key] = {'http_method': 'ANY', 'before_actions': []}

    self._routes[key]['function'] = func
    if route is not None:
      self._routes[key]['route'] = route
    if http_method is not None:
      self._routes[key]['http_method'] = http_method
    if jsonapi is not None:
      self._routes[key]['jsonapi'] = True
    if append_before_action is not None:
      self._routes[key]['before_actions'].append(append_before_action)

  @property
  def routes(self):
    return self._routes.values()


from functools import wraps
from .route_holder import RouteHolder

def route(route_string):
  def decorator_wrapper(func):
    holder = RouteHolder()
    holder.add_route_partial(func, route=route_string)

    @wraps(func)
    def func_wrapper(*args, **kwArgs):
      return func(*args, **kwArgs)
    return func_wrapper
  return decorator_wrapper



from functools import wraps
from .route_holder import RouteHolder

def jsonapi(reject_if_no_data=False):
  def decorator_wrapper(func):
    holder = RouteHolder()
    if reject_if_no_data:
      def check_no_data(req, respondent):
        if not req.data:
          print('JSON API Not parsed properly. Parsed Body is |{}|'.format(req.body))
          print('Original event is |{}|'.format(req.event))
          respondent.add_error({
            'code': 'invalid_jsonapi',
            'title': 'Body is not valid JSONAPI'
          })
      holder.add_route_partial(func, jsonapi=True, append_before_action=check_no_data)
    else:
      holder.add_route_partial(func, jsonapi=True)

    @wraps(func)
    def func_wrapper(*args, **kwArgs):
      return func(*args, **kwArgs)
    return func_wrapper
  return decorator_wrapper



import re

"""
# Path Construction

* No leading or trailing slashes
* Each element can either be a raw string, or a url-capture group
* URL capture groups are surround by {}
* At present, the router does *NOT* support slurping route components (`{proxy+}` in API Gateway, e.g.), except as the last parameter
"""

def parse_route_for_path(route_string, path):
  route_pieces = re.sub(r'/$', '', re.sub(r'^/', '', route_string.lower())).split('/')
  path_pieces = re.sub(r'/$', '', re.sub(r'^/', '', path.lower())).split('/')

  if len(route_pieces) > len(path_pieces):
    return None
  
  if not re.search(r'^\{.*\+\}$', route_pieces[-1]) and len(route_pieces) != len(path_pieces):
    return None

  path_params = {}
  for idx, route_elem in enumerate(route_pieces):
    match_obj = re.search(r'^\{(.*?)(\+*)\}$', route_elem)
    if match_obj:
      if match_obj.group(2) and idx == len(route_pieces) - 1: # Last element
        path_params[match_obj.group(1)] = '/'.join(path_pieces[idx:])
      else:
        path_params[match_obj.group(1)] = path_pieces[idx]
    else:
      if route_elem != path_pieces[idx]:
        return None
  return path_params


import json
import re
import json_api_doc
from .route_parser import parse_route_for_path

class EnhancedRequest:
  def __init__(self, event, route):
    self._event = event
    self._route = route
    self._body = {}
    self._data = {}

    params = {}
    if event['queryStringParameters']:
      params = {**params, **event['queryStringParameters']}
    if event['multiValueQueryStringParameters']:
      params = {**params, **event['multiValueQueryStringParameters']}
    params = {**params, **event['pathParameters']}
    if event['body']:
      params['raw_body'] = event['body']
      params['parsed_body'] = {}
      if re.search('json', event['headers'].get('content-type', ''), re.IGNORECASE):
        try:
          parsed_body = json.loads(event['body'])
          params['parsed_body'] = {**parsed_body}
          params = {**params, **parsed_body}
          self._body = parsed_body
          if route['jsonapi'] and 'data' in self._body:
            try:
              self._data = json_api_doc.deserialize(self._body)
            except AttributeError:
              self._data = {}
        except json.JSONDecodeError:
          pass

    params = {**params, **parse_route_for_path(route['route'], event['pathParameters'].get('proxy', ''))}
    self._params = params
  
  @property
  def params(self):
    return self._params

  @property
  def body(self):
    return self._body

  @property
  def data(self):
    return self._data

  @property
  def event(self):
    return self._event

  @property
  def route(self):
    return self._route


import json
from deepmerge import always_merger

class ResponseHandler:
  def __init__(self):
    self._status_code = 204
    self._response_body = ""
    self._headers = {}
    self._cookies = {}
    self._errors = []
    self._return_immediately = False

  @property
  def return_immediately(self):
    return self._return_immediately

  @property
  def has_errors(self):
    return len(self._errors) > 0
  
  @property
  def _set_cookie_headers(self):
    def format_cookie(kvpack):
      key, value_dict = kvpack
      pieces = ['{}={}'.format(key, value_dict['value'])]
      for opt in value_dict:
        if opt.lower() in ['expires', 'max-age', 'domain', 'path', 'samesite']:
          pieces.append('{}={}'.format(opt, value_dict[opt]))
        if opt.lower() in ['secure', 'httponly']:
          if value_dict[opt]:
            pieces.append(opt)
      return '; '.join(pieces)

    return list(map(format_cookie, self._cookies.items()))

  @property
  def status_code(self):
    if self._errors:
      return 400
    if self._response_body and self._status_code == 204:
      return 200
    return self._status_code

  @property
  def response(self):
    hsh = {
      "statusCode": self.status_code
    }
    if self._errors:
      hsh["body"] = ', '.join(map(lambda x: x.get('title', x.get('code', 'Unknown')), self._errors))
    elif self._response_body:
      hsh["body"] = self._response_body

    if self._headers:
      hsh["headers"] = self._headers
    if self._cookies:
      hsh["multiValueHeaders"] = {
        'Set-Cookie': self._set_cookie_headers
      }
    return hsh

  def set_status(self, status):
    if isinstance(status, int):
      self._status_code = status

  def set_body(self, body):
    if isinstance(body, dict):
      self.set_json(body)
    else:
      self._response_body = body

  def set_json(self, body):
    self._response_body = json.dumps(body)
    self._headers['content-type'] = 'application/json'

  def merge_json(self, body):
    cur_json = {}
    try:
      cur_json = json.loads(self._response_body)
    except json.JSONDecodeError:
      cur_json = {}
    new_json = always_merger.merge(cur_json, body)
    self.set_json(new_json)

  def add_error(self, error):
    self._errors.append(error)

  def set_cookie(self, key, value, **options):
    self._cookies[key] = {
      'value': value,
      **options
    }

  def set_return_immediately(self):
    self._return_immediately = True


class Singleton(type):
  _instances = {}
  def __call__(cls, *args, **kwargs):
    if cls not in cls._instances:
      cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
    return cls._instances[cls]